import pytest
from pytest import approx
import numpy as np
from tempfile import TemporaryDirectory
from sklearn.datasets import load_iris, make_multilabel_classification, load_diabetes
from skelm import LargeELMRegressor

pd = pytest.importorskip("pandas")  # tests if Pandas is installed
pytest.importorskip("pyarrow")
dask = pytest.importorskip("dask")


@pytest.fixture
def data_class():
    return load_iris(return_X_y=True)

@pytest.fixture
def data_ml():
    return make_multilabel_classification()

@pytest.fixture
def data_reg():
    return load_diabetes(return_X_y=True)


def test_Input_DifferentLengths_Raises():
    elm = LargeELMRegressor()
    with pytest.raises(ValueError):
        elm.fit(['a', 'b', 'c'], ['d'])


def test_Input_MultipleFiles(data_reg):
    X, y = data_reg
    with TemporaryDirectory() as data_dir:
        X_files = [data_dir + "/X_{}.parquet".format(i) for i in range(3)]
        y_files = [data_dir + "/y_{}.parquet".format(i) for i in range(3)]
        for i in range(3):
            pd.DataFrame(X[i::3], columns=[str(c) for c in range(X.shape[1])]).to_parquet(fname=X_files[i])
            pd.DataFrame(y[i::3], columns=['Class']).to_parquet(fname=y_files[i])

        elm = LargeELMRegressor(batch_size=10)
        elm.fit(X_files, y_files)
        y_hat = elm.predict(X_files)

        assert np.mean(y_hat != y) < 0.33



import pytest
from pytest import approx
import numpy as np
from skelm import ELMClassifier, ELMRegressor
from sklearn.datasets import load_iris, load_boston

@pytest.fixture
def data_class():
    return load_iris(return_X_y=True)


@pytest.fixture
def data_reg():
    return load_boston(return_X_y=True)


def test_Classifier_predict_ReturnsIntegerArray():
    X = np.array([[1], [2], [3]])
    Y = np.array([0, 0, 1])
    elm = ELMClassifier().fit(X, Y)
    Yh = elm.predict(X)
    assert isinstance(Yh, np.ndarray)
    assert Yh == approx(Yh.astype(np.int))


def test_Classifier_WrongNumberOfFeatures_RaisesError(data_class):
    X, T = data_class
    elm = ELMClassifier().fit(X, T)
    with pytest.raises(ValueError):
        elm.predict(X[:, 1:])


def test_Regressor_WrongNumberOfFeatures_RaisesError(data_reg):
    X, T = data_reg
    elm = ELMRegressor().fit(X, T)
    with pytest.raises(ValueError):
        elm.predict(X[:, 1:])


def test_Classifier_Multilabel(data_class):
    X, T = data_class
    Y = np.ones((T.shape[0], 2))
    Y[:, 0] = T
    elm = ELMClassifier()
    elm.fit(X, Y)


def test_Classifier_SetClasses_IgnoresOther(data_class):
    X, T = data_class
    elm = ELMClassifier(classes=[0, 1])
    Yh = elm.fit(X, T).predict(X)
    assert set(Yh) == {0, 1}


def test_Classifier_PartialFit(data_class):
    X, T = data_class
    elm0 = ELMClassifier(n_neurons=4, alpha=1, random_state=0)
    elm1 = ELMClassifier(n_neurons=4, alpha=1, random_state=0)

    elm0.fit(X, T)
    elm1.partial_fit(X[::2], T[::2])
    elm1.partial_fit(X[1::2], T[1::2])

    assert elm0.solver_.coef_ == approx(elm1.solver_.coef_)


def test_IterativeClassification_FeedClassesOneByOne(data_class):
    X, T = data_class
    elm = ELMClassifier(classes=[0, -1, -2], n_neurons=10, alpha=1)

    X0 = X[T == 0]
    Y0 = T[T == 0]
    elm.partial_fit(X0, Y0)

    X1 = X[T == 1]
    Y1 = T[T == 1]
    elm.partial_fit(X1, Y1, update_classes=True)

    X2 = X[T == 2]
    Y2 = T[T == 2]
    elm.partial_fit(X2, Y2, update_classes=True)

    Yh = elm.predict(X)
    assert set(Yh) == {0, 1, 2}


def test_IterativeSolver_SkipIntermediateSolution(data_class):
    X, T = data_class
    elm = ELMClassifier(classes=[0, 1, 2], n_neurons=10, alpha=1)

    X0 = X[T == 0]
    Y0 = T[T == 0]
    elm.partial_fit(X0, Y0, compute_output_weights=False)

    X1 = X[T == 1]
    Y1 = T[T == 1]
    elm.partial_fit(X1, Y1, compute_output_weights=False)

    X2 = X[T == 2]
    Y2 = T[T == 2]
    elm.partial_fit(X2, Y2)

    Yh = elm.predict(X)
    assert set(Yh) == {0, 1, 2}


def test_MultipleHiddenLayers(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[2, 3], ufunc=['tanh', 'sigm'],
                       density=[None, None], pairwise_metric=[None, None])
    elm.fit(X, Y)
    assert len(elm.hidden_layers_) == 2


def test_MultipleHiddenLayers_MoreCombinations(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[1, 1, 1, 1, 1],
                       ufunc=['relu', 'sigm', np.sin, None, None],
                       density=[None, 0.5, 0.8, None, None],
                       pairwise_metric=[None, None, None, 'l1', 'chebyshev'])
    elm.fit(X, Y)
    assert len(elm.hidden_layers_) == 5

def test_MultipleHL_DefaultValues(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[2, 3])
    elm.fit(X, Y)

def test_MultipleHL_Ufunc_SingleValue(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[2, 3], ufunc='sigm')
    elm.fit(X, Y)

def test_MultipleHL_Density_SingleValue(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[2, 3], density=0.7)
    elm.fit(X, Y)

def test_MultipleHL_Pairwise_SingleValue(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[2, 3], pairwise_metric='l2')
    elm.fit(X, Y)

def test_MultipleHL_WrongDimensions_Raises(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(n_neurons=[1, 2, 3, 4], ufunc=['relu', 'sigm'])
    with pytest.raises(ValueError):
        elm.fit(X, Y)

def test_RegularizationAlpha_NegativeValue_Raises(data_class):
    X, Y = data_class
    elm = ELMClassifier(alpha=-1)
    with pytest.raises(ValueError):
        elm.fit(X, Y)


import pytest
import numpy as np
from sklearn.datasets import load_iris, load_boston

from skelm import ELMClassifier, ELMRegressor

pd = pytest.importorskip("pandas")  # tests if Pandas is installed

@pytest.fixture
def data_class():
    X, y = load_iris(return_X_y=True)
    return pd.DataFrame(X), pd.DataFrame(y)

@pytest.fixture
def data_reg():
    X, y = load_boston(return_X_y=True)
    return pd.DataFrame(X), pd.DataFrame(y)


def test_Pandas_ActuallyUsesDataFrames(data_class, data_reg):
    X, y = data_class
    assert isinstance(X, pd.DataFrame)
    assert isinstance(y, pd.DataFrame)

    X, y = data_reg
    assert isinstance(X, pd.DataFrame)
    assert isinstance(y, pd.DataFrame)


def test_Classifier_FitOnPandas_ReturnsNumpy(data_class):
    X, Y = data_class
    elm = ELMClassifier().fit(X, Y)
    Yh = elm.predict(X)
    assert isinstance(Yh, np.ndarray)

def test_Classification_BetterThanNaive(data_class):
    X, Y = data_class
    elm = ELMClassifier(random_state=0).fit(X, Y)
    score = elm.score(X, Y)
    assert score > 0.33

def test_Regressor_FitOnPandas_ReturnsNumpy(data_reg):
    X, Y = data_reg
    elm = ELMRegressor().fit(X, Y)
    Yh = elm.predict(X)
    assert isinstance(Yh, np.ndarray)

def test_Regression_BetterThanNaive(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(random_state=0).fit(X, Y)
    r2score = elm.score(X, Y)
    assert r2score > 0.3



import pytest
import numpy as np
from scipy.special import expit as sigmoid
from sklearn.datasets import load_boston, load_iris
from sklearn.preprocessing import RobustScaler
from sklearn.random_projection import GaussianRandomProjection, SparseRandomProjection
from scipy.sparse import csc_matrix, csr_matrix, coo_matrix, lil_matrix

from skelm import ELMRegressor, ELMClassifier


c_X, c_y = load_iris(return_X_y=True)
r_X, r_y = load_boston(return_X_y=True)
r_X = RobustScaler().fit_transform(r_X)

params = [
    ((r_X, r_y), ELMRegressor),
    ((c_X, c_y), ELMClassifier),
    ((csc_matrix(r_X), r_y), ELMRegressor),
    ((csc_matrix(c_X), c_y), ELMClassifier),
    ((csr_matrix(r_X), r_y), ELMRegressor),
    ((csr_matrix(c_X), c_y), ELMClassifier),
    ((coo_matrix(r_X), r_y), ELMRegressor),
    ((coo_matrix(c_X), c_y), ELMClassifier),
    ((lil_matrix(r_X), r_y), ELMRegressor),
    ((lil_matrix(c_X), c_y), ELMClassifier),
]


@pytest.mark.parametrize("data,elm_model", params)
def test_Default_SetNumberOfNeurons(data, elm_model):
    elm5 = elm_model(n_neurons=5, random_state=0).fit(*data)
    elm50 = elm_model(n_neurons=50, random_state=0).fit(*data)
    score5 = elm5.score(*data)
    score50 = elm50.score(*data)
    assert score50 > score5
    assert score50 > 0.33


@pytest.mark.parametrize("data,elm_model", params)
def test_LinearPart_CanBeIncluded(data, elm_model):
    elm = elm_model(include_original_features=True, random_state=0).fit(*data)
    score = elm.score(*data)
    assert score > 0.33


@pytest.mark.parametrize("data,elm_model", params)
def test_LinearPart_AddsExtraFeatures(data, elm_model):
    X, y = data
    n_neurons_basic = elm_model().fit(X, y).coef_.shape[0]
    n_neurons_with_orig = elm_model(include_original_features=True).fit(X, y).coef_.shape[0]
    assert n_neurons_basic < n_neurons_with_orig


@pytest.mark.parametrize("data,elm_model", params)
def test_DefaultNeurons_UseGaussianRandomProjection(data, elm_model):
    elm = elm_model().fit(*data)
    assert isinstance(elm.hidden_layers_[0].projection_, GaussianRandomProjection)


@pytest.mark.parametrize("data,elm_model", params)
def test_SparseELM_UseSparseRandomProjection(data, elm_model):
    elm = elm_model(density=0.1).fit(*data)
    assert isinstance(elm.hidden_layers_[0].projection_, SparseRandomProjection)


@pytest.mark.parametrize("data,elm_model", params)
def test_Ufunc_Sigmoid(data, elm_model):
    elm = elm_model(ufunc="sigm").fit(*data)
    assert elm.hidden_layers_[0].ufunc_ is sigmoid


@pytest.mark.parametrize("data,elm_model", params)
def test_DefaultUfunc_Tanh(data, elm_model):
    elm_default = elm_model().fit(*data)
    elm_explicit = elm_model(ufunc="tanh").fit(*data)
    assert elm_default.hidden_layers_[0].ufunc_ is np.tanh
    assert elm_explicit.hidden_layers_[0].ufunc_ is np.tanh


@pytest.mark.parametrize("data,elm_model", params)
def test_Ufunc_WrongName_ReturnsValueError(data, elm_model):
    elm = elm_model(ufunc="UnIcOrN")
    with pytest.raises(ValueError):
        elm.fit(*data)


@pytest.mark.parametrize("data,elm_model", params)
def test_Ufunc_CustomLambdaFunction_Works(data, elm_model):
    relu = lambda x: np.maximum(x, 0)
    elm = elm_model(ufunc=relu).fit(*data)
    assert elm.hidden_layers_[0].ufunc_ is relu


@pytest.mark.parametrize("data,elm_model", params)
def test_Ufunc_NumpyUfunc_Works(data, elm_model):
    elm = elm_model(ufunc=np.sin).fit(*data)
    assert elm.hidden_layers_[0].ufunc_ is np.sin


@pytest.mark.parametrize("data,elm_model", params)
def test_PairwiseKernel_Works(data, elm_model):
    elm = elm_model(pairwise_metric="euclidean").fit(*data)
    assert hasattr(elm.hidden_layers_[0].projection_, "pairwise_metric")


@pytest.mark.parametrize("data,elm_model", params)
def test_PairwiseKernel_TooManyNeurons_StillWorks(data, elm_model):
    X, y = data
    elm = elm_model(n_neurons=3 * X.shape[0], pairwise_metric="euclidean")
    elm.fit(X, y)


@pytest.mark.parametrize("data,elm_model", params)
def test_PairwiseDistances_AllKinds_FromScikitLearn(data, elm_model):
    elm_model(n_neurons=3, pairwise_metric="cityblock").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="cosine").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="euclidean").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="l1").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="l2").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="manhattan").fit(*data)


@pytest.mark.parametrize("data,elm_model", params)
def test_PairwiseDistances_AllKinds_FromScipy(data, elm_model):
    elm_model(n_neurons=3, pairwise_metric="braycurtis").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="canberra").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="chebyshev").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="correlation").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="dice").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="hamming").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="jaccard").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="kulsinski").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="mahalanobis").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="minkowski").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="rogerstanimoto").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="russellrao").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="seuclidean").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="sokalmichener").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="sokalsneath").fit(*data)
    elm_model(n_neurons=3, pairwise_metric="sqeuclidean").fit(*data)





import pytest
from sklearn.utils.estimator_checks import check_estimator
from skelm import ELMRegressor, ELMClassifier, BatchCholeskySolver, PairwiseRandomProjection


class MultiTaskELMRegressor(ELMRegressor):
    pass

class MultiTaskELMClassifier(ELMClassifier):
    pass


@pytest.mark.parametrize("method", [MultiTaskELMRegressor, MultiTaskELMClassifier, BatchCholeskySolver, PairwiseRandomProjection])
def test_DenseData_ScikitLearnCompatibility(method):
    check_estimator(method)




import scipy as sp
from enum import Enum
from sklearn.metrics import pairwise_distances
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_random_state


class HiddenLayerType(Enum):
    RANDOM = 1    # Gaussian random projection
    SPARSE = 2    # Sparse Random Projection
    PAIRWISE = 3  # Pairwise kernel with a number of centroids


def dummy(x):
    return x


def flatten(items):
    """Yield items from any nested iterable."""
    for x in items:
        # don't break strings into characters
        if hasattr(x, '__iter__') and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def _is_list_of_strings(obj):
    return obj is not None and all(isinstance(elem, str) for elem in obj)


def _dense(X):
    if sp.sparse.issparse(X):
        return X.todense()
    else:
        return X


class PairwiseRandomProjection(BaseEstimator, TransformerMixin):

    def __init__(self, n_components=100, pairwise_metric='l2', n_jobs=None, random_state=None):
        """Pairwise distances projection with random centroids.

        Parameters
        ----------
        n_components : int
            Number of components (centroids) in the projection. Creates the same number of output features.

        pairwise_metric : str
            A valid pairwise distance metric, see pairwise-distances_.
            .. _pairwise-distances: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html#sklearn.metrics.pairwise_distances

        n_jobs : int or None, optional, default=None
            Number of jobs to use in distance computations, or `None` for no parallelism.
            Passed to _pairwise-distances function.

        random_state
            Used for random generation of centroids.
        """
        self.n_components = n_components
        self.pairwise_metric = pairwise_metric
        self.n_jobs = n_jobs
        self.random_state = random_state

    def fit(self, X, y=None):
        """Generate artificial centroids.

        Centroids are sampled from a normal distribution. They work best if the data is normalized.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=[n_samples, n_features]
            Input data
        """
        X = check_array(X, accept_sparse=True)
        self.random_state_ = check_random_state(self.random_state)

        if self.n_components <= 0:
            raise ValueError("n_components must be greater than 0, got %s" % self.n_components)

        self.components_ = self.random_state_.randn(self.n_components, X.shape[1])
        self.n_jobs_ = 1 if self.n_jobs is None else self.n_jobs
        return self

    def transform(self, X):
        """Compute distance matrix between input data and the centroids.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            Input data samples.

        Returns
        -------
        X_dist : numpy array
            Distance matrix between input data samples and centroids.
        """
        X = check_array(X, accept_sparse=True)
        check_is_fitted(self, 'components_')

        if X.shape[1] != self.components_.shape[1]:
            raise ValueError(
                'Impossible to perform projection: X at fit stage had a different number of features. '
                '(%s != %s)' % (X.shape[1], self.components_.shape[1]))

        try:
            X_dist = pairwise_distances(X, self.components_, n_jobs=self.n_jobs_, metric=self.pairwise_metric)
        except TypeError:
            # scipy distances that don't support sparse matrices
            X_dist = pairwise_distances(_dense(X), _dense(self.components_), n_jobs=self.n_jobs_, metric=self.pairwise_metric)

        return X_dist


import pytest
import tempfile
from sklearn.datasets import load_boston, load_iris
from sklearn.preprocessing import RobustScaler
from scipy.sparse import csc_matrix, csr_matrix, coo_matrix, lil_matrix
import pickle

from skelm import ELMRegressor, ELMClassifier


@pytest.fixture
def data_class():
    return load_iris(return_X_y=True)


@pytest.fixture
def data_reg():
    return load_boston(return_X_y=True)


def test_Serialize_Solver(data_reg):
    X, Y = data_reg
    elm = ELMRegressor(random_state=0)
    elm.fit(X, Y)
    Yh1 = elm.predict(X)

    solver_data = pickle.dumps(elm.solver_, pickle.HIGHEST_PROTOCOL)
    del elm.solver_

    elm.solver_ = pickle.loads(solver_data)
    Yh2 = elm.predict(X)
    assert Yh1 == pytest.approx(Yh2)


def test_Serialize_HiddenLayer(data_class):
    X, Y = data_class
    elm = ELMClassifier(
        n_neurons=(5,6,7), ufunc=('tanh', None, 'sigm'), density=(None, None, 0.5),
        pairwise_metric=(None, 'l1', None), random_state=0)
    elm.fit(X, Y)
    Yh1 = elm.predict(X)

    hl_data = [pickle.dumps(hl, pickle.HIGHEST_PROTOCOL) for hl in elm.hidden_layers_]
    del elm.hidden_layers_

    elm.hidden_layers_ = [pickle.loads(z) for z in hl_data]
    Yh2 = elm.predict(X)
    assert Yh1 == pytest.approx(Yh2)


def test_Serialize_ELM(data_class):
    X, Y = data_class
    elm = ELMClassifier(
        n_neurons=(5,6,7), ufunc=('tanh', None, 'sigm'), density=(None, None, 0.5),
        pairwise_metric=(None, 'l1', None), random_state=0)
    elm.fit(X, Y)
    Yh1 = elm.predict(X)

    elm_data = pickle.dumps(elm, pickle.HIGHEST_PROTOCOL)
    elm2 = pickle.loads(elm_data)

    Yh2 = elm.predict(X)
    assert Yh1 == pytest.approx(Yh2)


def test_Serialize_ContinueTraining(data_class):
    X, Y = data_class
    x1, y1 = X[0::2], Y[0::2]
    x2, y2 = X[1::2], Y[1::2]

    elm1 = ELMClassifier(n_neurons=10, random_state=0)
    elm1.fit(X, Y)
    Yh1 = elm1.predict(X)

    elm2 = ELMClassifier(n_neurons=10, random_state=0)
    elm2.fit(x1, y1)
    elm2_data = pickle.dumps(elm2, pickle.HIGHEST_PROTOCOL)
    del elm2
    elm2_loaded = pickle.loads(elm2_data)
    elm2_loaded.partial_fit(x2, y2)
    Yh2 = elm2_loaded.predict(X)

    assert Yh1 == pytest.approx(Yh2)


def test_SerializeToFile_ContinueTraining(data_class):
    X, Y = data_class
    x1, y1 = X[0::2], Y[0::2]
    x2, y2 = X[1::2], Y[1::2]

    elm1 = ELMClassifier(n_neurons=10, random_state=0)
    elm1.fit(X, Y)
    Yh1 = elm1.predict(X)

    elm2 = ELMClassifier(n_neurons=10, random_state=0)
    elm2.fit(x1, y1)
    with tempfile.TemporaryFile() as ftemp:
        pickle.dump(elm2, ftemp, pickle.HIGHEST_PROTOCOL)
        del elm2
        ftemp.seek(0)
        elm2_reloaded = pickle.load(ftemp)

    elm2_reloaded.partial_fit(x2, y2)
    Yh2 = elm2_reloaded.predict(X)

    assert Yh1 == pytest.approx(Yh2)


import numpy as np
from sklearn.base import RegressorMixin
from sklearn.metrics import pairwise_distances
from sklearn.utils.validation import check_is_fitted, check_array

import dask.array as da
import dask.dataframe as dd

from .elm import _BaseELM
from dask.distributed import Client, LocalCluster, wait
from .utils import _is_list_of_strings, _dense, HiddenLayerType, dummy


def _read_numeric_file(fname):
    try:
        return dd.read_parquet(fname)
    except:
        pass

    try:
        return dd.read_csv(fname)
    except:
        pass

    try:
        return np.load(fname)
    except:
        pass

class LargeELMRegressor(_BaseELM, RegressorMixin):
    """ELM Regressor for larger-than-memory problems.

    Uses `Dask <https://dask.org>`_ for batch analysis of data in Parquet files.

    .. attention:: Why do I need Parquet files?

        Parquet files provide necessary information about the data without loading whole file content from
        disk. It makes a tremendous runtime difference compared to simpler `.csv` or `.json` file formats.
        Reading from files saves memory by loading data in small chunks, supporting arbitrary large input files.
        It also solves current memory leaks with Numpy matrix inputs in Dask.

        Any data format can be easily converted to Parquet, see `Analytical methods <techniques.html>`_ section.

        HDF5 is almost as good as Parquet, but performs worse with Dask due to internal data layout.

    .. todo: Write converters.

    .. todo: Memo about number of workers: one is good, several cover disk read latency but need more memory.
        On one machine matrix operators always run in parallel, do not benefit from Dask.

    .. todo: Memory consumption with large number of neurons - 100,000 neurons require 200GB or swap space, with
        read+write reaching 1GB/s. Suggested a fast SSD, or HDD + extra workers to hide swap latency.
        Mention that Dask is not the perfect solution, kept here for future updates. And it actually solves
        stuff larger than memory, albeit at a very high time+swap cost.

    .. todo: Avoid large batch sizes as workers can fail, safe bet is 2000-5000 range.

    .. todo: Fast HtH and in-place Cholesky solver.

    .. todo: Pro tip in documentation: run ELM with dummy 1000 data samples and 1e+9 regularization,
        This will test possible memory issues for workers without wasting your time on computing full HH.

    .. todo: Option to keep full HH permanently somewhere at disk. Saves before the final step,
        avoids failures from memory issues during Cholesky solver.

    .. todo: GPU + batch Cholesky solver, for both ELM and LargeELM.

    Requirements
    ------------
        * Pandas
        * pyarrow
        * python-snappy

    Parameters
    ----------

    batch_size : int
        Batch size used for both data samples and hidden neurons. With batch Cholesky solver, allows for very large
        numbers of hidden neurons of over 100,000; limited only by the computation time and disk swap space.

        .. hint:: Include bias and original features for best performance.

        ELM will include a bias term (1 extra feature), and the original features with `include_original_features=True`.
        For optimal performance, choose `batch_size` to be equal or evenly divide the
        `n_neurons + 1 (bias) + n_inputs (if include_original_features=True)`.

        .. todo:: Exact batch_size vs. GPU performance
    """


    def __del__(self):
        if hasattr(self, 'client_'):
            self.client_.close()
            self.cluster_.close()

    def _setup_dask_client(self):
        self.cluster_ = LocalCluster(
            n_workers=4, threads_per_worker=1,
            local_dir="/Users/akusok/wrkdir/dask-temp",
            memory_limit="8GB"
        )
        self.client_ = Client(self.cluster_)

        W_list = [hl.projection_.components_ for hl in self.hidden_layers_]
        W_dask = [da.from_array(_dense(W), chunks=self.bsize_) for W in W_list]
        self.W_ = self.client_.persist(W_dask)

        def foo():
            import os
            os.environ['OMP_NUM_THREADS'] = '1'
        self.client_.run(foo)

        print("Running on:", self.client_)

        try:
            dashboard = self.client_.scheduler_info()['address'].split(":")
            dashboard[0] = "http"
            dashboard[-1] = str(self.client_.scheduler_info()['services']['dashboard'])
            print("Dashboard at", ":".join(dashboard))
        except:
            pass

    def _project(self, X_dask):
        """Compute hidden layer output with Dask functionality.
        """
        H_list = []
        for hl, W in zip(self.hidden_layers_, self.W_):
            if hl.hidden_layer_ == HiddenLayerType.PAIRWISE:
                H0 = X_dask.map_blocks(
                    pairwise_distances,
                    W,
                    dtype=X_dask.dtype,
                    chunks=(X_dask.chunks[0], (W.shape[0],)),
                    metric=hl.pairwise_metric
                )
            else:
                XW_dask = da.dot(X_dask, W.transpose())
                if hl.ufunc_ is dummy:
                    H0 = XW_dask
                elif hl.ufunc_ is np.tanh:
                    H0 = da.tanh(XW_dask)
                else:
                    H0 = XW_dask.map_blocks(hl.ufunc_)
            H_list.append(H0)

        if self.include_original_features:
            H_list.append(X_dask)
        H_list.append(da.ones((X_dask.shape[0], 1)))

        H_dask = da.concatenate(H_list, axis=1).rechunk(self.bsize_)
        return H_dask

    def _compute(self, X, y, sync_every, HH=None, HY=None):
        """Computing matrices HH and HY, the actually long part.

        .. todo: actually distributed computations that scatter batches of data file names,
            and reduce-sum the HH,HY matrices.
        """

        # processing files
        for i, X_file, y_file in zip(range(len(X)), X, y):
            X_dask = dd.read_parquet(X_file).to_dask_array(lengths=True)
            Y_dask = dd.read_parquet(y_file).to_dask_array(lengths=True)
            H_dask = self._project(X_dask)

            if HH is None:  # first iteration
                HH = da.dot(H_dask.transpose(), H_dask)
                HY = da.dot(H_dask.transpose(), Y_dask)
            else:
                HH += da.dot(H_dask.transpose(), H_dask)
                HY += da.dot(H_dask.transpose(), Y_dask)
                if sync_every is not None and i % sync_every == 0:
                    wait([HH, HY])

            # synchronization
            if sync_every is not None and i % sync_every == 0:
                HH, HY = self.client_.persist([HH, HY])

        # finishing solution
        if sync_every is not None:
            wait([HH, HY])
        return HH, HY

    def _solve(self, HH, HY):
        """Compute output weights from HH and HY using Dask functionality.
        """
        # make HH/HY divisible by chunk size
        n_features, _ = HH.shape
        padding = 0
        if n_features > self.bsize_ and n_features % self.bsize_ > 0:
            print("Adjusting batch size {} to n_features {}".format(self.bsize_, n_features))
            padding = self.bsize_ - (n_features % self.bsize_)
            P01 = da.zeros((n_features, padding))
            P10 = da.zeros((padding, n_features))
            P11 = da.zeros((padding, padding))
            HH = da.block([[HH,  P01],
                           [P10, P11]])

            P1 = da.zeros((padding, HY.shape[1]))
            HY = da.block([[HY],
                           [P1]])

        # rechunk, add bias, and solve
        HH = HH.rechunk(self.bsize_) + self.alpha * da.eye(HH.shape[1], chunks=self.bsize_)
        HY = HY.rechunk(self.bsize_)

        B = da.linalg.solve(HH, HY, sym_pos=True)
        if padding > 0:
            B = B[:n_features]

        return B

    def fit(self, X, y=None, sync_every=10):
        """Fits an ELM with data in a bunch of files.

        Model will use the set of features from the first file.
        Same features must have same names across the whole dataset.

        .. todo: Check what happens if features are in different order or missing.

        Does **not** support sparse data.

        .. todo: Check if some sparse data would work.

        .. todo: Check that sync_every does not affect results

        .. todo: Add single precision

        .. todo: Parquet file format examples in documentation

        Original features and bias are added to the end of data, for easier rechunk-merge. This way full chunks
        of hidden neuron outputs stay intact.


        Parameters
        ----------

        X : [str]
            List of input data files in Parquet format.

        y : [str]
            List of target data files in Parquet format.

        sync_every : int or None
            Synchronize computations after this many files are processed. None for running without synchronization.
            Less synchronization improves run speed with smaller data files, but may result in large swap space usage
            for large data problems. Use smaller number for more frequent synchronization if swap space
            becomes a problem.
        """

        if not _is_list_of_strings(X) or not _is_list_of_strings(y):
            raise ValueError("Expected X and y as lists of file names.")

        if len(X) != len(y):
            raise ValueError("Expected X and y as lists of files with the same length. "
                             "Got len(X)={} and len(y)={}".format(len(X), len(y)))

        # read first file and get parameters
        X_dask = dd.read_parquet(X[0]).to_dask_array(lengths=True)
        Y_dask = dd.read_parquet(y[0]).to_dask_array(lengths=True)

        n_samples, n_features = X_dask.shape
        if hasattr(self, 'n_features_') and self.n_features_ != n_features:
            raise ValueError('Shape of input is different from what was seen in `fit`')

        _, n_outputs = Y_dask.shape
        if hasattr(self, 'n_outputs_') and self.n_outputs_ != n_outputs:
            raise ValueError('Shape of outputs is different from what was seen in `fit`')

        # set batch size, default is bsize=2000 or all-at-once with less than 10_000 samples
        self.bsize_ = self.batch_size
        if self.bsize_ is None:
            self.bsize_ = n_samples if n_samples < 10 * 1000 else 2000

        # init model if not fit yet
        if not hasattr(self, 'hidden_layers_'):
            self.n_features_ = n_features
            self.n_outputs_ = n_outputs

            X_sample = X_dask[:10].compute()
            self._init_hidden_layers(X_sample)
            self._setup_dask_client()

        HH, HY = self._compute(X, y, sync_every=sync_every)
        self.B = self._solve(HH, HY)
        self.is_fitted_ = True
        return self

    def predict(self, X):
        """Prediction works with both lists of Parquet files and numeric arrays.

        Parameters
        ----------

        X : array-like, [str]
            Input data as list of Parquet files, or as a numeric array.

        Returns
        -------
        Yh : array, shape (n_samples, n_outputs)
            Predicted values for all input samples.

            .. attention:: Returns all outputs as a single in-memory array!

                Danger of running out out memory for high-dimensional outputs, if a large set of input
                files is provided. Feed data in smaller batches in such case.
        """
        check_is_fitted(self, 'is_fitted_')

        if _is_list_of_strings(X):
            Yh_list = []

            # processing files
            for X_file in X:
                X_dask = dd.read_parquet(X_file).to_dask_array(lengths=True)
                H_dask = self._project(X_dask)
                Yh_list.append(da.dot(H_dask, self.B))

            Yh_dask = da.concatenate(Yh_list, axis=0)
            return Yh_dask.compute()

        else:
            X = check_array(X, accept_sparse=True)
            H = [np.ones((X.shape[0], 1))]
            if self.include_original_features:
                H.append(_dense(X))
            H.extend([hl.transform(X) for hl in self.hidden_layers_])

            return np.hstack(H) @ self.B.compute()



import pytest
from sklearn.datasets import load_iris, load_boston
from skelm import ELMClassifier, ELMRegressor

@pytest.fixture
def data_class():
    return load_iris(return_X_y=True)

@pytest.fixture
def data_reg():
    return load_boston(return_X_y=True)


def test_Classification_Iris_BetterThanNaive(data_class):
    elm = ELMClassifier(random_state=0)
    elm.fit(*data_class)
    score = elm.score(*data_class)
    assert score > 0.33


def test_Regression_Boston_BetterThanNaive(data_reg):
    elm = ELMRegressor(random_state=0)
    elm.fit(*data_reg)
    r2score = elm.score(*data_reg)
    assert r2score > 0.3




import pytest
from pytest import approx
import numpy as np
from sklearn.datasets import load_iris, make_multilabel_classification, load_diabetes
from skelm import ELMRegressor, ELMClassifier


@pytest.fixture
def data_class():
    return load_iris(return_X_y=True)

@pytest.fixture
def data_ml():
    return make_multilabel_classification()

@pytest.fixture
def data_reg():
    return load_diabetes(return_X_y=True)

def test_SineWave_Solves():
    """A highly non-linear regression problem, with added strong noise.
    """
    X = np.linspace(-1, 1, num=1000)[:, None]
    Y = np.sin(16 * X) * X + 0.2*np.random.randn(1000)[:, None]

    elm = ELMRegressor(random_state=0)
    elm.fit(X, Y)
    Yt = elm.predict(X)

    MSE = np.mean((Y - Yt) ** 2)
    assert MSE < 0.3


def test_Xor_OneNeuron_Solved():
    """ELM should be able to solve XOR problem.
    """
    X = np.array([[0, 0],
                  [1, 1],
                  [1, 0],
                  [0, 1]])
    Y = np.array([1, 1, -1, -1])

    elm = ELMClassifier(n_neurons=3, random_state=0)
    elm.fit(X, Y)
    Yh = elm.predict(X)
    assert Yh[0] > 0
    assert Yh[1] > 0
    assert Yh[2] < 0
    assert Yh[3] < 0


def test_ELMClassifier_ReportedScore_ActuallyIsClassificationScore(data_class):
    X, Y = data_class
    Yr = np.vstack((Y == 0, Y == 1, Y == 2)).T

    elm_c = ELMClassifier(random_state=0).fit(X, Y)
    elm_r = ELMRegressor(random_state=0).fit(X, Yr)

    Yc_hat = elm_c.predict(X)
    Yr_hat = elm_r.predict(X).argmax(1)

    assert Yc_hat == approx(Yr_hat)


def test_ELMClassifier_MultilabelClassification_Works(data_ml):
    X, Y = data_ml
    elm_c = ELMClassifier(random_state=0).fit(X, Y)
    elm_r = ELMRegressor(random_state=0).fit(X, Y)

    Yc_hat = elm_c.predict(X)
    Yr_hat = (elm_r.predict(X) >= 0.5).astype(np.int)

    assert Yc_hat == approx(Yr_hat)


def test_RegularizationL2_DifferentValue_ChangesPrediction(data_reg):
    X, Y = data_reg
    Yh_1 = ELMRegressor(alpha=1e-7, random_state=0).fit(X, Y).predict(X)
    Yh_2 = ELMRegressor(alpha=1e+3, random_state=0).fit(X, Y).predict(X)

    assert Yh_1 != approx(Yh_2)



import numpy as np
import scipy as sp
import warnings
from sklearn.exceptions import DataConversionWarning
from sklearn.base import BaseEstimator, RegressorMixin

from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils import check_X_y, check_array

from dask import distributed
from dask.distributed import Client, LocalCluster
import dask.dataframe as dd
import dask.array as da


class DaskCholeskySolver(BaseEstimator, RegressorMixin):
    """Out-of-core linear system solver with Dask back-end.

    Parameters
    ----------

    alpha : float, non-negative
        L2 regularization parameter, larger value means stronger effect. The value may be
        increased if the system fails to converge; actual used value stored in `alpha_` parameter.

    batch_size : int
        Batch size for **samples and features**. Computations proceed on square blocks of data.
        For optimal performance, use a number of features that is equal or a bit less than multiple
        of a batch size; e.g. 8912 features with 3000 batch size.

    swap_dir : str
        Directory for temporary storage of Dask data that does not fit in memory. A large and fast
        storage is advised, like a local SSD.


    Attributes
    ----------

    cluster_ : object
        An instance of `dask.distributed.LocalCluster`.

    client_ : object
        Dask client for running computations.
    """

    def __init__(self, alpha=1e-7, batch_size=2000, swap_dir=None):
        self.alpha = alpha
        self.batch_size = batch_size
        self.swap_dir = swap_dir

    def _init_dask(self):
        self.cluster_ = LocalCluster( n_workers=2, local_dir=self.swap_dir)
        self.client_ = Client(self.cluster_)
        print("Running on:")
        print(self.client_)

    def fit(self, X, y):


        self.W_ = da.random.normal

        return self

    def predict(self, X):
        return None


class BBvdsnjvlsdnjhbgfndjvksdjkvlndsf(BaseEstimator, RegressorMixin):

    def __init__(self, alpha=1e-7):
        self.alpha = alpha

    def _init_XY(self, X, y):
        """Initialize covariance matrices, including a separate bias term.
        """
        d_in = X.shape[1]
        self._XtX = np.eye(d_in + 1) * self.alpha
        self._XtX[0, 0] = 0
        if len(y.shape) == 1:
            self._XtY = np.zeros((d_in + 1,))
        else:
            self._XtY = np.zeros((d_in + 1, y.shape[1]))

    @property
    def XtY_(self):
        return self._XtY

    @property
    def XtX_(self):
        return self._XtX

    @XtY_.setter
    def XtY_(self, value):
        self._XtY = value

    @XtX_.setter
    def XtX_(self, value):
        self._XtX = value

    def _solve(self):
        """Second stage of solution (X'X)B = X'Y using Cholesky decomposition.

        Sets `is_fitted_` to True.
        """
        B = sp.linalg.solve(self._XtX, self._XtY, assume_a='pos', overwrite_a=False, overwrite_b=False)
        self.coef_ = B[1:]
        self.intercept_ = B[0]
        self.is_fitted_ = True

    def _reset(self):
        """Erase solution and data matrices.
        """
        [delattr(self, attr) for attr in ('_XtX', '_XtY', 'coef_', 'intercept_', 'is_fitted_') if hasattr(self, attr)]

    def fit(self, X, y):
        """Solves an L2-regularized linear system like Ridge regression, overwrites any previous solutions.
        """
        self._reset()  # remove old solution
        self.partial_fit(X, y, compute_output_weights=True)
        return self

    def partial_fit(self, X, y, compute_output_weights=True):
        """Update model with a new batch of data.

        Output weight computation can be temporary turned off for faster processing. This will mark model as
        not fit. Enable `compute_output_weights` in the final call to `partial_fit`.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=[n_samples, n_features]
            Training input samples

        y : array-like, shape=[n_samples, n_targets]
            Training targets

        compute_output_weights : boolean, optional, default True
            Whether to compute new output weights (coef_, intercept_). Disable this in intermediate `partial_fit`
            steps to run computations faster, then enable in the last call to compute the new solution.

            .. Note::
                Solution can be updated without extra data by setting `X=None` and `y=None`.
        """
        if self.alpha < 0:
            raise ValueError("Regularization parameter alpha must be non-negative.")

        # solution only
        if X is None and y is None and compute_output_weights:
            self._solve()
            return self

        # validate parameters
        X, y = check_X_y(X, y, accept_sparse=True, multi_output=True, y_numeric=True, ensure_2d=True)
        if len(y.shape) > 1 and y.shape[1] == 1:
            msg = "A column-vector y was passed when a 1d array was expected.\
                   Please change the shape of y to (n_samples, ), for example using ravel()."
            warnings.warn(msg, DataConversionWarning)

        # init temporary data storage
        if not hasattr(self, '_XtX'):
            self._init_XY(X, y)
        else:
            if X.shape[1] + 1 != self._XtX.shape[0]:
                n_new, n_old = X.shape[1], self._XtX.shape[0] - 1
                raise ValueError("Number of features %d does not match previous data %d." % (n_new, n_old))

        # compute temporary data
        X_sum = safe_sparse_dot(X.T, np.ones((X.shape[0],)))
        y_sum = safe_sparse_dot(y.T, np.ones((y.shape[0],)))
        self._XtX[0, 0] += X.shape[0]
        self._XtX[1:, 0] += X_sum
        self._XtX[0, 1:] += X_sum
        self._XtX[1:, 1:] += X.T @ X

        self._XtY[0] += y_sum
        self._XtY[1:] += X.T @ y

        # solve
        if not compute_output_weights:
            # mark as not fitted
            [delattr(self, attr) for attr in ('coef_', 'intercept_', 'is_fitted_') if hasattr(self, attr)]
        else:
            self._solve()
        return self

    def predict(self, X):
        check_is_fitted(self, 'is_fitted_')
        X = check_array(X, accept_sparse=True)
        return safe_sparse_dot(X, self.coef_, dense_output=True) + self.intercept_



import pytest
from pytest import approx
import numpy as np

from skelm.solver_batch import BatchCholeskySolver


def test_SingleStepSolution():
    X = np.random.randn(100, 3)
    Y = X @ np.array([1, 2, 3]) - 2
    solver = BatchCholeskySolver().fit(X[::2], Y[::2])
    Yh = solver.predict(X[1::2])
    assert Y[1::2] == approx(Yh)
    assert solver.coef_ == approx(np.array([1, 2, 3]), rel=1e-3)
    assert solver.intercept_ == approx(-2)


def test_PartialFitSolution():
    X = np.random.randn(100, 3)
    Y = X @ np.array([1, 2, 3]) - 2
    solver = BatchCholeskySolver().partial_fit(X[::2], Y[::2])
    Yh = solver.predict(X[1::2])
    assert Y[1::2] == approx(Yh, rel=1e-3)
    assert solver.coef_ == approx(np.array([1, 2, 3]), rel=1e-3)
    assert solver.intercept_ == approx(-2)


def test_OutputShape_1d():
    X = np.random.randn(100, 3)
    Y = X @ np.array([1, 2, 3]) - 2
    assert len(Y.shape) == 1
    solver = BatchCholeskySolver().partial_fit(X[::2], Y[::2])
    Yh = solver.predict(X[1::2])
    assert len(Yh.shape) == 1


def test_OutputShape_2d():
    X = np.random.randn(100, 3)
    W = np.array([[1, 4],
                  [2, 5],
                  [3, 6]])
    Y = X @ W - 2
    assert len(Y.shape) == 2
    solver = BatchCholeskySolver().partial_fit(X[::2], Y[::2])
    Yh = solver.predict(X[1::2])
    assert len(Yh.shape) == 2


def test_PartialFit_SeveralParts():
    X = np.random.randn(100, 3)
    W = np.array([[1, 4],
                  [2, 5],
                  [3, 6]])
    Y = X @ W - 2

    solver = BatchCholeskySolver()
    
    # give 1st and 2nd output column separately, multiply by 2
    # to get same results as if both columns were present both times
    solver.partial_fit(X[::2], 2 * Y[::2] @ np.array([[1, 0], [0, 0]]))
    solver.partial_fit(X[::2], 2 * Y[::2] @ np.array([[0, 0], [0, 1]]))

    Yh = solver.predict(X[1::2])
    assert Y[1::2] == approx(Yh, rel=1e-3)
    assert solver.coef_ == approx(W, rel=1e-3)
    assert solver.intercept_ == approx(np.array([-2, -2]))




"""
High-level Extreme Learning Machine modules
"""

import numpy as np
import warnings
from scipy.special import expit

from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, clone
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels, type_of_target

from sklearn.preprocessing import LabelBinarizer, MultiLabelBinarizer
from sklearn.exceptions import DataConversionWarning, DataDimensionalityWarning

from .hidden_layer import HiddenLayer
from .solver_batch import BatchCholeskySolver
from .utils import _dense

warnings.simplefilter("ignore", DataDimensionalityWarning)


class _BaseELM(BaseEstimator):

    def __init__(self, alpha=1e-7, batch_size=None, include_original_features=False,
                 n_neurons=None, ufunc="tanh", density=None, pairwise_metric=None,
                 random_state=None):

        self.alpha = alpha
        self.n_neurons = n_neurons
        self.batch_size = batch_size
        self.ufunc = ufunc
        self.include_original_features = include_original_features
        self.density = density
        self.pairwise_metric = pairwise_metric
        self.random_state = random_state

    def _init_hidden_layers(self, X):
        """Init an empty model, creating objects for hidden layers and solver.

        Also validates inputs for several hidden layers.
        """
        # only one type of neurons
        if not hasattr(self.n_neurons, '__iter__'):
            hl = HiddenLayer(n_neurons=self.n_neurons, density=self.density, ufunc=self.ufunc,
                             pairwise_metric=self.pairwise_metric, random_state=self.random_state)
            hl.fit(X)
            self.hidden_layers_ = (hl, )

        # several different types of neurons
        else:
            k = len(self.n_neurons)

            # fix default values
            ufuncs = self.ufunc
            if isinstance(ufuncs, str) or not hasattr(ufuncs, "__iter__"):
                ufuncs = [ufuncs] * k

            densities = self.density
            if densities is None or not hasattr(densities, "__iter__"):
                densities = [densities] * k

            pw_metrics = self.pairwise_metric
            if pw_metrics is None or isinstance(pw_metrics, str):
                pw_metrics = [pw_metrics] * k

            if not k == len(ufuncs) == len(densities) == len(pw_metrics):
                raise ValueError("Inconsistent parameter lengths for model with {} different types of neurons.\n"
                                 "Set 'ufunc', 'density' and 'pairwise_distances' by lists "
                                 "with {} elements, or leave the default values.".format(k, k))

            self.hidden_layers_ = []
            for n_neurons, ufunc, density, metric in zip(self.n_neurons, ufuncs, densities, pw_metrics):
                hl = HiddenLayer(n_neurons=n_neurons, density=density, ufunc=ufunc,
                                 pairwise_metric=metric, random_state=self.random_state)
                hl.fit(X)
                self.hidden_layers_.append(hl)

    def _reset(self):
        [delattr(self, attr) for attr in ('n_features_', 'solver_', 'hidden_layers_', 'is_fitted_', 'label_binarizer_') if hasattr(self, attr)]

    @property
    def n_neurons_(self):
        if not hasattr(self, 'hidden_layers_'):
            return None

        neurons_count = sum([hl.n_neurons_ for hl in self.hidden_layers_])
        if self.include_original_features:
            neurons_count += self.n_features_

        return neurons_count

    @property
    def coef_(self):
        return self.solver_.coef_

    @property
    def intercept_(self):
        return self.solver_.intercept_

    def partial_fit(self, X, y=None, forget=False, compute_output_weights=True):
        """Update model with a new batch of data.

        |method_partial_fit|

        .. |method_partial_fit| replace:: Output weight computation can be temporary turned off
            for faster processing. This will mark model as not fit. Enable `compute_output_weights`
            in the final call to `partial_fit`.

        .. |param_forget| replace:: Performs a negative update, effectively removing the information
            given by training samples from the model. Output weights need to be re-computed after forgetting
            data. Forgetting data that have not been learned before leads to unpredictable results.

        .. |param_compute_output_weights| replace::  Whether to compute new output weights
            (coef_, intercept_). Disable this in intermediate `partial_fit`
            steps to run computations faster, then enable in the last call to compute the new solution.


        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=[n_samples, n_features]
            Training input samples

        y : array-like, shape=[n_samples, n_targets]
            Training targets

        forget : boolean, default False
            |param_forget|

        compute_output_weights : boolean, optional, default True
            |param_compute_output_weights|

            .. Note::
                Solution can be updated without extra data by setting `X=None` and `y=None`.

            Example:
                >>> model.partial_fit(X_1, y_1)
                ... model.partial_fit(X_2, y_2)
                ... model.partial_fit(X_3, y_3)    # doctest: +SKIP

            Faster, option 1:
                >>> model.partial_fit(X_1, y_1, compute_output_weights=False)
                ... model.partial_fit(X_2, y_2, compute_output_weights=False)
                ... model.partial_fit(X_3, y_3)    # doctest: +SKIP

            Faster, option 2:
                >>> model.partial_fit(X_1, y_1, compute_output_weights=False)
                ... model.partial_fit(X_2, y_2, compute_output_weights=False)
                ... model.partial_fit(X_3, y_3, compute_output_weights=False)
                ... model.partial_fit(X=None, y=None)    # doctest: +SKIP
        """
        # compute output weights only
        if X is None and y is None and compute_output_weights:
            self.solver_.partial_fit(None, None, compute_output_weights=True)
            self.is_fitted_ = True
            return self

        X, y = check_X_y(X, y, accept_sparse=True, multi_output=True)
        if len(y.shape) > 1 and y.shape[1] == 1:
            msg = ("A column-vector y was passed when a 1d array was expected. "
                   "Please change the shape of y to (n_samples, ), for example using ravel().")
            warnings.warn(msg, DataConversionWarning)

        n_samples, n_features = X.shape
        if hasattr(self, 'n_features_') and self.n_features_ != n_features:
            raise ValueError('Shape of input is different from what was seen in `fit`')

        # set batch size, default is bsize=2000 or all-at-once with less than 10_000 samples
        self.bsize_ = self.batch_size
        if self.bsize_ is None:
            self.bsize_ = n_samples if n_samples < 10 * 1000 else 2000

        # init model if not fit yet
        if not hasattr(self, 'hidden_layers_'):
            self.n_features_ = n_features
            self.solver_ = BatchCholeskySolver(alpha=self.alpha)
            self._init_hidden_layers(X)

        # special case of one-shot processing
        if self.bsize_ >= n_samples:
            H = [hl.transform(X) for hl in self.hidden_layers_]
            H = np.hstack(H if not self.include_original_features else [_dense(X)] + H)
            self.solver_.partial_fit(H, y, forget=forget, compute_output_weights=False)

        else:  # batch processing
            for b_start in range(0, n_samples, self.bsize_):
                b_end = min(b_start + self.bsize_, n_samples)
                b_X = X[b_start:b_end]
                b_y = y[b_start:b_end]

                b_H = [hl.transform(b_X) for hl in self.hidden_layers_]
                b_H = np.hstack(b_H if not self.include_original_features else [_dense(b_X)] + b_H)
                self.solver_.partial_fit(b_H, b_y, forget=forget, compute_output_weights=False)

        # output weights if needed
        if compute_output_weights:
            self.solver_.partial_fit(None, None, compute_output_weights=True)
            self.is_fitted_ = True

        # mark as needing a solution
        elif hasattr(self, 'is_fitted_'):
            del self.is_fitted_

        return self

    def fit(self, X, y=None):
        """Reset model and fit on the given data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Training data samples.

        y : array-like, shape (n_samples,) or (n_samples, n_outputs)
            Target values used as real numbers.

        Returns
        -------
        self : object
            Returns self.
        """

        #todo: add X as bunch of files support

        self._reset()
        self.partial_fit(X, y)
        return self

    def predict(self, X):
        """Predict real valued outputs for new inputs X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Input data samples.

        Returns
        -------
        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            Predicted outputs for inputs X.

            .. attention::

                :mod:`predict` always returns a dense matrix of predicted outputs -- unlike
                in :meth:`fit`, this may cause memory issues at high number of outputs
                and very high number of samples. Feed data by smaller batches in such case.
        """

        X = check_array(X, accept_sparse=True)
        check_is_fitted(self, "is_fitted_")

        H = [hl.transform(X) for hl in self.hidden_layers_]
        if self.include_original_features:
            H = [_dense(X)] + H
        H = np.hstack(H)

        return self.solver_.predict(H)


class ELMRegressor(_BaseELM, RegressorMixin):
    """Extreme Learning Machine for regression problems.

    This model solves a regression problem, that is a problem of predicting continuous outputs.
    It supports multi-variate regression (when ``y`` is a 2d array of shape [n_samples, n_targets].)
    ELM uses ``L2`` regularization, and optionally includes the original data features to
    capture linear dependencies in the data natively.

    Parameters
    ----------
    alpha : float
        Regularization strength; must be a positive float. Larger values specify stronger effect.
        Regularization improves model stability and reduces over-fitting at the cost of some learning
        capacity. The same value is used for all targets in multi-variate regression.

        The optimal regularization strength is suggested to select from a large range of logarithmically
        distributed values, e.g. :math:`[10^{-5}, 10^{-4}, 10^{-3}, ..., 10^4, 10^5]`. A small default
        regularization value of :math:`10^{-7}` should always be present to counter numerical instabilities
        in the solution; it does not affect overall model performance.

        .. attention::
            The model may automatically increase the regularization value if the solution
            becomes unfeasible otherwise. The actual used value contains in ``alpha_`` attribute.

    batch_size : int, optional
        Actual computations will proceed in batches of this size, except the last batch that may be smaller.
        Default behavior is to process all data at once with <10,000 samples, otherwise use batches
        of size 2000.

    include_original_features : boolean, default=False
        Adds extra hidden layer neurons that simpy copy the input data features, adding a linear part
        to the final model solution that can directly capture linear relations between data and
        outputs. Effectively increases `n_neurons` by `n_inputs` leading to a larger model.
        Including original features is generally a good thing if the number of data features is low.

    n_neurons : int or [int], optional
        Number of hidden layer neurons in ELM model, controls model size and learning capacity.
        Generally number of neurons should be less than the number of training data samples, as
        otherwise the model will learn the training set perfectly resulting in overfitting.

        Several different kinds of neurons can be used in the same model by specifying a list of
        neuron counts. ELM will create a separate neuron type for each element in the list.
        In that case, the following attributes ``ufunc``, ``density`` and ``pairwise_metric``
        should be lists of the same length; default values will be automatically expanded into a list.

        .. note::
            Models with <1,000 neurons are very fast to compute, while GPU acceleration is efficient
            starting from 1,000-2,000 neurons. A standard computer should handle up to 10,000 neurons.
            Very large models will not fit in memory but can still be trained by an out-of-core solver.

    ufunc : {'tanh', 'sigm', 'relu', 'lin' or callable}, or a list of those (see n_neurons)
        Transformation function of hidden layer neurons. Includes the following options:
            - 'tanh' for hyperbolic tangent
            - 'sigm' for sigmoid
            - 'relu' for rectified linear unit (clamps negative values to zero)
            - 'lin' for linear neurons, transformation function does nothing
            - any custom callable function like members of ``Numpu.ufunc``

    density : float in range (0, 1], or a list of those (see n_neurons), optional
        Specifying density replaces dense projection layer by a sparse one with the specified
        density of the connections. For instance, ``density=0.1`` means each hidden neuron will
        be connected to a random 10% of input features. Useful for working on very high-dimensional
        data, or for large numbers of neurons.

    pairwise_metric : {'euclidean', 'cityblock', 'cosine' or other}, or a list of those (see n_neurons), optional
        Specifying pairwise metric replaces multiplicative hidden neurons by distance-based hidden
        neurons. This ELM model is known as Radial Basis Function ELM (RBF-ELM).

        .. note::
            Pairwise function neurons ignore ufunc and density.

        Typical metrics are `euclidean`, `cityblock` and `cosine`. For a full list of metrics check
        the `webpage <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html>`_
        of :mod:`sklearn.metrics.pairwise_distances`.

    random_state : int, RandomState instance or None, optional, default None
        The seed of the pseudo random number generator to use when generating random numbers e.g.
        for hidden neuron parameters. Random state instance is passed to lower level objects and routines.
        Use it for repeatable experiments.

    Attributes
    ----------
    n_neurons_ : int
        Number of automatically generated neurons.

    ufunc_ : function
        Tranformation function of hidden neurons.

    projection_ : object
        Hidden layer projection function.

    solver_ : object
        Solver instance, read solution from there.


    Examples
    --------

    Combining ten sigmoid and twenty RBF neurons in one model:

    >>> model = ELMRegressor(n_neurons=(10, 20),
    ...                      ufunc=('sigm', None),
    ...                      density=(None, None),
    ...                      pairwise_metric=(None, 'euclidean'))   # doctest: +SKIP

    Default values in multi-neuron ELM are automatically expanded to a list

    >>>  model = ELMRegressor(n_neurons=(10, 20),
    ...                       ufunc=('sigm', None),
    ...                       pairwise_metric=(None, 'euclidean'))   # doctest: +SKIP

    >>>  model = ELMRegressor(n_neurons=(30, 30),
    ...                       pairwise_metric=('cityblock', 'cosine'))   # doctest: +SKIP
    """
    pass

class ELMClassifier(_BaseELM, ClassifierMixin):
    """ELM classifier, modified for multi-label classification support.

    :param classes: Set of classes to consider in the model; can be expanded at runtime.
                    Samples of other classes will have their output set to zero.
    :param solver: Solver to use, "default" for build-in Least Squares or "ridge" for Ridge regression


    Example descr...

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`fit`.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """

    def __init__(self, classes=None, alpha=1e-7, batch_size=None, include_original_features=False, n_neurons=None,
                 ufunc="tanh", density=None, pairwise_metric=None, random_state=None):
        super().__init__(alpha, batch_size, include_original_features, n_neurons, ufunc, density, pairwise_metric,
                         random_state)
        self.classes = classes

    @property
    def classes_(self):
        return self.label_binarizer_.classes_

    def _get_tags(self):
        return {"multioutput": True, "multilabel": True}

    def _update_classes(self, y):
        if not isinstance(self.solver_, BatchCholeskySolver):
            raise ValueError("Only iterative solver supports dynamic class update")

        old_classes = self.label_binarizer_.classes_
        partial_classes = clone(self.label_binarizer_).fit(y).classes_

        # no new classes detected
        if set(partial_classes) <= set(old_classes):
            return

        if len(old_classes) < 3:
            raise ValueError("Dynamic class update has to start with at least 3 classes to function correctly; "
                             "provide 3 or more 'classes=[...]' during initialization.")

        # get new classes sorted by LabelBinarizer
        self.label_binarizer_.fit(np.hstack((old_classes, partial_classes)))
        new_classes = self.label_binarizer_.classes_

        # convert existing XtY matrix to new classes
        if hasattr(self.solver_, 'XtY_'):
            XtY_old = self.solver_.XtY_
            XtY_new = np.zeros((XtY_old.shape[0], new_classes.shape[0]))
            for i, c in enumerate(old_classes):
                j = np.where(new_classes == c)[0][0]
                XtY_new[:, j] = XtY_old[:, i]
            self.solver_.XtY_ = XtY_new

        # reset the solution
        if hasattr(self.solver_, 'is_fitted_'):
            del self.solver_.is_fitted_

    def partial_fit(self, X, y=None, forget=False, update_classes=False, compute_output_weights=True):
        """Update classifier with a new batch of data.

        |method_partial_fit|

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape=[n_samples, n_features]
            Training input samples

        y : array-like, shape=[n_samples, n_targets]
            Training targets

        forget : boolean, default False
            |param_forget|

        update_classes : boolean, default False
            Include new classes from `y` into the model, assuming they were 0 in all previous samples.

        compute_output_weights : boolean, optional, default True
            |param_compute_output_weights|
        """

        #todo: Warning on strongly non-normalized data

        X, y = check_X_y(X, y, accept_sparse=True, multi_output=True)

        # init label binarizer if needed
        if not hasattr(self, 'label_binarizer_'):
            self.label_binarizer_ = LabelBinarizer()
            if type_of_target(y).endswith("-multioutput"):
                self.label_binarizer_ = MultiLabelBinarizer()
            self.label_binarizer_.fit(self.classes if self.classes is not None else y)

        if update_classes:
            self._update_classes(y)

        y_numeric = self.label_binarizer_.transform(y)
        if len(y_numeric.shape) > 1 and y_numeric.shape[1] == 1:
            y_numeric = y_numeric[:, 0]

        super().partial_fit(X, y_numeric, forget=forget, compute_output_weights=compute_output_weights)
        return self

    def fit(self, X, y=None):
        """Fit a classifier erasing any previously trained model.

        Returns
        -------
        self : object
            Returns self.
        """
        self._reset()
        self.partial_fit(X, y, compute_output_weights=True)
        return self

    def predict(self, X):
        """Predict classes of new inputs X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,) or (n_samples, n_outputs)
            Returns one most probable class for multi-class problem, or
            a binary vector of all relevant classes for multi-label problem.
        """

        check_is_fitted(self, "is_fitted_")
        scores = super().predict(X)
        return self.label_binarizer_.inverse_transform(scores)

    def predict_proba(self, X):
        """Probability estimation for all classes.

        Positive class probabilities are computed as
        1. / (1. + np.exp(-self.decision_function(X)));
        multiclass is handled by normalizing that over all classes.
        """
        check_is_fitted(self, "is_fitted_")
        prob = super().predict(X)
        expit(prob, out=prob)
        if prob.ndim == 1:
            return np.vstack([1 - prob, prob]).T
        else:
            # OvR normalization, like LibLinear's predict_probability
            prob /= prob.sum(axis=1).reshape((prob.shape[0], -1))
            return prob


import numpy as np
import scipy as sp

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_array, check_is_fitted

from sklearn.random_projection import GaussianRandomProjection, SparseRandomProjection
from .utils import PairwiseRandomProjection, HiddenLayerType, dummy

# suppress annoying warning of random projection into a higher-dimensional space
import warnings
warnings.filterwarnings("ignore", message="DataDimensionalityWarning")


def auto_neuron_count(n, d):
    # computes default number of neurons for `n` data samples with `d` features
    return min(int(250 * np.log(1 + d/10) - 15), n//3 + 1)


ufuncs = {"tanh": np.tanh,
          "sigm": sp.special.expit,
          "relu": lambda x: np.maximum(x, 0),
          "lin": dummy,
          None: dummy}


class HiddenLayer(BaseEstimator, TransformerMixin):

    def __init__(self, n_neurons=None, density=None, ufunc="tanh", pairwise_metric=None, random_state=None):
        self.n_neurons = n_neurons
        self.density = density
        self.ufunc = ufunc
        self.pairwise_metric = pairwise_metric
        self.random_state = random_state

    def _fit_random_projection(self, X):
        self.hidden_layer_ = HiddenLayerType.RANDOM
        self.projection_ = GaussianRandomProjection(n_components=self.n_neurons_, random_state=self.random_state_)
        self.projection_.fit(X)

    def _fit_sparse_projection(self, X):
        self.hidden_layer_ = HiddenLayerType.SPARSE
        self.projection_ = SparseRandomProjection(n_components=self.n_neurons_, density=self.density,
                                                  dense_output=True, random_state=self.random_state_)
        self.projection_.fit(X)

    def _fit_pairwise_projection(self, X):
        self.hidden_layer_ = HiddenLayerType.PAIRWISE
        self.projection_ = PairwiseRandomProjection(n_components=self.n_neurons_,
                                                    pairwise_metric=self.pairwise_metric,
                                                    random_state=self.random_state_)
        self.projection_.fit(X)

    def fit(self, X, y=None):
        # basic checks
        X = check_array(X, accept_sparse=True)

        # handle random state
        self.random_state_ = check_random_state(self.random_state)

        # get number of neurons
        n, d = X.shape
        self.n_neurons_ = int(self.n_neurons) if self.n_neurons is not None else auto_neuron_count(n, d)

        # fit a projection
        if self.pairwise_metric is not None:
            self._fit_pairwise_projection(X)
        elif self.density is not None:
            self._fit_sparse_projection(X)
        else:
            self._fit_random_projection(X)

        if self.ufunc in ufuncs.keys():
            self.ufunc_ = ufuncs[self.ufunc]
        elif callable(self.ufunc):
            self.ufunc_ = self.ufunc
        else:
            raise ValueError("Ufunc transformation function not understood: ", self.ufunc)

        self.is_fitted_ = True
        return self

    def transform(self, X):
        check_is_fitted(self, "is_fitted_")

        X = check_array(X, accept_sparse=True)
        n_features = self.projection_.components_.shape[1]
        if X.shape[1] != n_features:
            raise ValueError("X has %d features per sample; expecting %d" % (X.shape[1], n_features))

        if self.hidden_layer_ == HiddenLayerType.PAIRWISE:
            return self.projection_.transform(X)  # pairwise projection ignores ufunc

        return self.ufunc_(self.projection_.transform(X))



import pytest
from sklearn.datasets import load_iris, load_boston
from skelm import ELMClassifier, ELMRegressor


@pytest.mark.parametrize(
    "data,elm", [(load_iris(return_X_y=True), ELMClassifier), (load_boston(return_X_y=True), ELMRegressor)]
)
def test_NeuronsAmount_CanSetManually(data, elm):
    elm_3 = elm(n_neurons=3, random_state=0)
    elm_3.fit(*data)
    score_3 = elm_3.score(*data)

    elm_10 = elm(n_neurons=10, random_state=0)
    elm_10.fit(*data)
    score_10 = elm_10.score(*data)

    assert score_10 > score_3

@pytest.mark.parametrize(
    "data,elm", [(load_iris(return_X_y=True), ELMClassifier), (load_boston(return_X_y=True), ELMRegressor)]
)
def test_Forgetting_SameResults(data, elm):
    X, y = data
    X1, X2, X3 = [X[i::3] for i in range(3)]
    y1, y2, y3 = [y[i::3] for i in range(3)]

    elm1 = elm(n_neurons=10, random_state=0)
    elm1.fit(X1, y1)
    score1 = elm1.score(X3, y3)

    elm2 = elm(n_neurons=10, random_state=0)
    elm2.fit(X, y)
    elm2.partial_fit(X2, y2, forget=True)
    elm2.partial_fit(X3, y3, forget=True)
    score2 = elm2.score(X3, y3)

    assert pytest.approx(score1) == score2



from .elm import ELMRegressor
from .elm import ELMClassifier
from .large_elm import LargeELMRegressor
from .hidden_layer import HiddenLayer
from .solver_batch import BatchCholeskySolver
from .utils import PairwiseRandomProjection

from ._version import __version__

__all__ = ['ELMRegressor', 'ELMClassifier', 'HiddenLayer',
           'LargeELMRegressor',
           'BatchCholeskySolver', 'PairwiseRandomProjection', '__version__']
